import json
import logging
import math
from dataclasses import dataclass
from json.decoder import JSONDecodeError
from typing import Literal
from typing import Optional
from typing import Union

import requests

logger = logging.getLogger(__name__)
ethgastypes = Literal["fast", "fastest", "safeLow", "average", "standard"]


@dataclass
class GasStation:
    api: str
    parse_rsp: list[Union[str, int, ethgastypes]]


async def fetch_gas_price() -> Optional[int]:
    """Estimate current ETH gas price

    Current implementation fetches from ethgasstation

    Returns:
        eth gas price in gwei
    """
    return await ethgasstation("fast")


async def ethgasstation(style: ethgastypes = "fast", retries: int = 2) -> Optional[int]:
    """Fetch gas price from ethgasstation in gwei"""
    for _ in range(retries):
        try:
            rsp = requests.get("https://ethgasstation.info/json/ethgasAPI.json")
            prices = json.loads(rsp.content)
            gas_price = int(prices[style])
            return int(gas_price / 10)  # json output is gwei*10
        except JSONDecodeError:
            logger.error("Error decoding JSON from ethgasstation API")
            continue
        except requests.exceptions.SSLError:
            logger.error("SSLError -- Unable to fetch gas price")
            return None
        except Exception as e:
            logger.error(f"Error fetching gas price: {e}")
    return None


ETH_GAS_PRICE_API = "https://ethgasstation.info/json/ethgasAPI.json"
MATIC_GAS_PRICE_API = "https://gasstation-mainnet.matic.network"
CHIADO_GAS_PRICE_API = "https://blockscout.com/xdai/mainnet/api/v1/gas-price-oracle"
GNOSIS_GAS_PRICE_API = "https://blockscout.com/xdai/mainnet/api/v1/gas-price-oracle"
OPTIMISM_GAS_PRICE_API = "https://api.owlracle.info/v3/opt/gas"
ARBITRUM_GAS_PRICE_API = "https://api.owlracle.info/v3/arb/gas"

gas_station = {
    1: GasStation(api=ETH_GAS_PRICE_API, parse_rsp=["fast"]),
    5: GasStation(api=ETH_GAS_PRICE_API, parse_rsp=["fast"]),
    11155111: GasStation(api=ETH_GAS_PRICE_API, parse_rsp=["fast"]),
    10: GasStation(api=OPTIMISM_GAS_PRICE_API, parse_rsp=["speeds", 2, "gasPrice"]),
    420: GasStation(api=OPTIMISM_GAS_PRICE_API, parse_rsp=["speeds", 2, "gasPrice"]),
    42161: GasStation(api=ARBITRUM_GAS_PRICE_API, parse_rsp=["speeds", 2, "gasPrice"]),
    421613: GasStation(api=ARBITRUM_GAS_PRICE_API, parse_rsp=["speeds", 2, "gasPrice"]),
    137: GasStation(api=MATIC_GAS_PRICE_API, parse_rsp=["safeLow"]),
    80001: GasStation(api=MATIC_GAS_PRICE_API, parse_rsp=["safeLow"]),
    10200: GasStation(api=CHIADO_GAS_PRICE_API, parse_rsp=["average"]),
    100: GasStation(api=GNOSIS_GAS_PRICE_API, parse_rsp=["average"]),
}


async def legacy_gas_station(
    chain_id: int, speed_parse_lis: Optional[list[Union[str, int, ethgastypes]]] = None, retries: int = 2
) -> Optional[int]:
    """Fetch gas price from gas station Api in gwei"""
    prices = {}
    if chain_id not in gas_station:
        logger.error(f"Please add gas station API for chain id: {chain_id}")
        return None

    for _ in range(retries):
        try:
            rsp = requests.get(gas_station[chain_id].api)
            prices = json.loads(rsp.content)
        except JSONDecodeError:
            logger.error("Error decoding JSON from gasstation API")
            continue
        except requests.exceptions.SSLError:
            logger.error("SSLError: Unable to fetch gas price")
            return None
        except Exception as e:
            logger.error(f"Error fetching gas price: {e}")
            return None

    if speed_parse_lis is None:
        speed_parse_lis = gas_station[chain_id].parse_rsp

    for i in speed_parse_lis:
        try:
            prices = prices[i]
        except (KeyError, IndexError):
            logger.error(f"Unable to parse gas price from gasstation: {speed_parse_lis}")
            return None

    if isinstance(prices, int):
        gas_price = prices
    elif isinstance(prices, float):
        gas_price = int(prices) if prices > 1 else math.ceil(prices)
    else:
        logger.error(f"Invalid reponse from gas station API: {prices}")
        return None

    return gas_price if chain_id not in (1, 5) else int(gas_price / 10)  # json output is gwei*10 for eth


if __name__ == "__main__":
    import asyncio

    loop = asyncio.get_event_loop()
    for chain_id in gas_station:
        price = loop.run_until_complete(legacy_gas_station(chain_id))
        assert isinstance(price, int)
        assert price > 0
        print(chain_id, price)



"""
Test covering gas price fetching utils
"""
import pytest
from requests.exceptions import SSLError

from src.telliot_core.gas.legacy_gas import fetch_gas_price
from src.telliot_core.gas.legacy_gas import legacy_gas_station


def raise_ssl_error(*args, **kwargs):
    raise SSLError


def raise_exception():
    raise Exception


@pytest.mark.asyncio
async def test_fetch_legacy_gas_price(monkeypatch, caplog):

    monkeypatch.setattr("requests.get", lambda x: raise_ssl_error())
    gp = await fetch_gas_price()

    assert gp is None
    assert "SSLError -- Unable to fetch gas price" in caplog.text

    monkeypatch.setattr("requests.get", lambda x: raise_exception())
    gp = await fetch_gas_price()

    assert gp is None
    assert "Error fetching gas price:" in caplog.text


@pytest.mark.asyncio
async def test_legacy_gasstation(monkeypatch, caplog):
    """Test legacy_gasstation"""

    gp = await legacy_gas_station(chain_id=0)

    assert gp is None
    assert "Please add gas station API for chain id: 0" in caplog.text

    gp = await legacy_gas_station(chain_id=1, speed_parse_lis="premium")

    assert gp is None
    assert "Unable to parse gas price from gasstation: premium" in caplog.text

    monkeypatch.setattr("requests.get", lambda x: raise_ssl_error())
    gp = await legacy_gas_station(chain_id=1)

    assert gp is None
    assert "SSLError: Unable to fetch gas price" in caplog.text

    monkeypatch.setattr("requests.get", lambda x: raise_exception())
    gp = await legacy_gas_station(chain_id=1)

    assert gp is None
    assert "Error fetching gas price:" in caplog.text



from dataclasses import dataclass
from dataclasses import field
from typing import List
from typing import Optional

from telliot_core.apps.config import ConfigOptions
from telliot_core.model.base import Base


@dataclass
class EVMCurrency(Base):
    name: str
    symbol: str
    decimals: int


@dataclass
class Chain(Base):
    name: str
    chain: str
    network: str
    chain_id: int
    currency: EVMCurrency


default_chain_list = [
    Chain(
        chain_id=1,
        name="Ethereum Mainnet",
        chain="ETH",
        network="mainnet",
        currency=EVMCurrency(name="Ether", symbol="ETH", decimals=18),
    ),
    Chain(
        chain_id=3,
        name="Ethereum Testnet Ropsten",
        chain="ETH",
        network="ropsten",
        currency=EVMCurrency(name="Ropsten Ether", symbol="ROP", decimals=18),
    ),
    Chain(
        chain_id=4,
        name="Ethereum Testnet Rinkeby",
        chain="ETH",
        network="rinkeby",
        currency=EVMCurrency(name="Rinkeby Ether", symbol="RIN", decimals=18),
    ),
    Chain(
        chain_id=137,
        name="Matic(Polygon) Mainnet",
        chain="Matic(Polygon)",
        network="mainnet",
        currency=EVMCurrency(name="Matic", symbol="MATIC", decimals=18),
    ),
    Chain(
        chain_id=69,
        name="Optimism Kovan",
        chain="Optimism",
        network="testnet (public)",
        currency=EVMCurrency(name="Ether", symbol="ETH", decimals=18),
    ),
    Chain(
        chain_id=1666600000,
        name="Harmony Mainnet",
        chain="Harmony",
        network="mainnet",
        currency=EVMCurrency(name="Harmony ONE", symbol="ONE", decimals=18),
    ),
    Chain(
        chain_id=1666700000,
        name="Harmony Testnet",
        chain="Harmony",
        network="testnet",
        currency=EVMCurrency(name="Harmony ONE", symbol="ONE", decimals=18),
    ),
    Chain(
        chain_id=421611,
        name="Arbitrum Rinkeby",
        chain="Arbitrum",
        network="testnet",
        currency=EVMCurrency(name="Ether", symbol="ETH", decimals=18),
    ),
    Chain(
        chain_id=42161,
        name="Arbitrum One",
        chain="Arbitrum",
        network="mainnet",
        currency=EVMCurrency(name="Arbitrum Ether", symbol="ETH", decimals=18),
    ),
    Chain(
        chain_id=421613,
        name="Arbitrum Goerli",
        chain="Arbitrum-goerli",
        network="testnet-goerli",
        currency=EVMCurrency(name="Arbitrum Goerli Ether", symbol="AGOR", decimals=18),
    ),
    Chain(
        chain_id=10200,
        name="Chiado Testnet",
        chain="Chiado",
        network="testnet",
        currency=EVMCurrency(name="Chiado xDai", symbol="XDAI", decimals=18),
    ),
    Chain(
        chain_id=100,
        name="gnosis",
        chain="gnosis",
        network="mainnet",
        currency=EVMCurrency(name="xDai", symbol="XDAI", decimals=18),
    ),
    Chain(
        chain_id=10,
        name="optimism",
        chain="optimism",
        network="mainnet",
        currency=EVMCurrency(name="Optimism Ether", symbol="ETH", decimals=18),
    ),
    Chain(
        chain_id=420,
        name="optimism-goerli",
        chain="optimism-goerli",
        network="testnet-goerli",
        currency=EVMCurrency(name="Optimism Ether", symbol="ETH", decimals=18),
    ),
    Chain(
        chain_id=3141,
        name="filecoin-hyperspace",
        chain="filecoin-hyperspace",
        network="testnet",
        currency=EVMCurrency(name="Testnet Filecoin", symbol="FIL", decimals=18),
    ),
    Chain(
        chain_id=314159,
        name="filecoin-calibration",
        chain="filecoin-calibration",
        network="testnet",
        currency=EVMCurrency(name="Testnet Filecoin", symbol="FIL", decimals=18),
    ),
    Chain(
        chain_id=314,
        name="filecoin",
        chain="filecoin",
        network="mainnet",
        currency=EVMCurrency(name="Filecoin", symbol="FIL", decimals=18),
    ),
    Chain(
        chain_id=11155111,
        name="Ethereum Testnet Sepolia",
        chain="sepolia",
        network="testnet",
        currency=EVMCurrency(name="Ether", symbol="ETH", decimals=18),
    ),
    Chain(
        chain_id=369,
        name="pulsechain",
        chain="pulsechain",
        network="mainnet",
        currency=EVMCurrency(name="PulseChain PLS", symbol="PLS", decimals=18),
    ),
    Chain(
        chain_id=943,
        name="pulsechain-testnet",
        chain="pulsechain-testnet",
        network="testnet",
        currency=EVMCurrency(name="PulseChain tPLS", symbol="tPLS", decimals=18),
    ),
    Chain(
        chain_id=3441005,
        name="manta-testnet",
        chain="manta-testnet",
        network="testnet",
        currency=EVMCurrency(name="Ether", symbol="ETH", decimals=18),
    ),
    Chain(
        chain_id=84531,
        name="base-goerli",
        chain="base-goerli",
        network="testnet",
        currency=EVMCurrency(name="Ether", symbol="ETH", decimals=18),
    ),
    Chain(
        chain_id=5001,
        name="mantle-testnet",
        chain="mantle-testnet",
        network="testnet",
        currency=EVMCurrency(name="Mantle", symbol="MNT", decimals=18),
    ),
    Chain(
        chain_id=5000,
        name="mantle",
        chain="mantle",
        network="mainnet",
        currency=EVMCurrency(name="Mantle", symbol="MNT", decimals=18),
    ),
    Chain(
        chain_id=2442,
        name="Polygon zkEVM Cardona Testnet",
        chain="cardona",
        network="testnet",
        currency=EVMCurrency(name="Ether", symbol="ETH", decimals=18),
    ),
    Chain(
        chain_id=1101,
        name="Polygon zkEVM",
        chain="zkEVM",
        network="mainnet",
        currency=EVMCurrency(name="Ether", symbol="ETH", decimals=18),
    ),
    Chain(
        chain_id=59140,
        name="Linea Goerli",
        chain="Linea Goerli",
        network="testnet",
        currency=EVMCurrency(name="Ether", symbol="ETH", decimals=18),
    ),
    Chain(
        chain_id=59144,
        name="Linea",
        chain="Linea",
        network="mainnet",
        currency=EVMCurrency(name="Ether", symbol="ETH", decimals=18),
    ),
    Chain(
        chain_id=2522,
        name="Fraxtal Testnet",
        chain="Fraxtal Testnet",
        network="testnet",
        currency=EVMCurrency(name="Frax Ether", symbol="frxETH", decimals=18),
    ),
    Chain(
        chain_id=252,
        name="Fraxtal Mainnet",
        chain="Fraxtal Mainnet",
        network="mainnet",
        currency=EVMCurrency(name="Frax Ether", symbol="frxETH", decimals=18),
    ),
    Chain(
        chain_id=1998,
        name="Kyoto Testnet",
        chain="Kyoto Testnet",
        network="testnet",
        currency=EVMCurrency(name="KYOTO", symbol="KYOTO", decimals=18),
    ),
    Chain(
        chain_id=1444673419,
        name="Skale Europa Testnet",
        chain="Skale Europa Testnet",
        network="testnet",
        currency=EVMCurrency(name="Skale Fuel", symbol="sFUEL", decimals=18),
    ),
    Chain(
        chain_id=2046399126,
        name="Skale Europa Mainnet",
        chain="Skale Europa Mainnet",
        network="mainnet",
        currency=EVMCurrency(name="Skale Fuel", symbol="sFUEL", decimals=18),
    ),
    Chain(
        chain_id=59141,
        name="Linea Sepolia",
        chain="Linea Sepolia",
        network="testnet",
        currency=EVMCurrency(name="Ether", symbol="ETH", decimals=18),
    ),
    Chain(
        chain_id=324,
        name="zkSync Era Mainnet",
        chain="zkSync Era Mainnet",
        network="network",
        currency=EVMCurrency(name="Ether", symbol="ETH", decimals=18),
    ),
    Chain(
        chain_id=300,
        name="zkSync Era Sepolia Testnet",
        chain="zkSync Era Sepolia Testnet",
        network="testnet",
        currency=EVMCurrency(name="Ether", symbol="ETH", decimals=18),
    ),
    Chain(
        chain_id=80002,
        name="Polygon Amoy Testnet",
        chain="polygon-amoy",
        network="testnet",
        currency=EVMCurrency(name="Matic", symbol="MATIC", decimals=18),
    ),
    Chain(
        chain_id=11155420,
        name="Optimism Sepolia Testnet",
        chain="optimism-sepolia",
        network="testnet",
        currency=EVMCurrency(name="Ether", symbol="ETH", decimals=18),
    ),
    Chain(
        chain_id=421614,
        name="Arbitrum Sepolia Testnet",
        chain="arbitrum-sepolia",
        network="testnet",
        currency=EVMCurrency(name="Ether", symbol="ETH", decimals=18),
    ),
    Chain(
        chain_id=5003,
        name="Mantle Sepolia Testnet",
        chain="mantle-sepolia",
        network="testnet",
        currency=EVMCurrency(name="Mantle", symbol="MNT", decimals=18),
    ),
    Chain(
        chain_id=84532,
        name="Base Sepolia Testnet",
        chain="base-sepolia",
        network="testnet",
        currency=EVMCurrency(name="Ether", symbol="ETH", decimals=18),
    ),
    Chain(
        chain_id=111,
        name="puff-bob-jznbxtoq7h",
        chain="BOB",
        network="testnet",
        currency=EVMCurrency(name="Ether", symbol="ETH", decimals=18),
    ),
    Chain(
        chain_id=60808,
        name="BOB",
        chain="BOB",
        network="mainnet",
        currency=EVMCurrency(name="Ether", symbol="ETH", decimals=18),
    ),
    Chain(
        chain_id=919,
        name="mode-sepolia-vtnhnpim72",
        chain="mode",
        network="testnet",
        currency=EVMCurrency(name="Ether", symbol="ETH", decimals=18),
    ),
    Chain(
        chain_id=1918988905,
        name="Rari Testnet",
        chain="rari",
        network="testnet",
        currency=EVMCurrency(name="Ether", symbol="ETH", decimals=18),
    ),
    Chain(
        chain_id=41,
        name="Telos EVM Testnet",
        chain="telos",
        network="testnet",
        currency=EVMCurrency(name="Telos", symbol="TLOS", decimals=18),
    ),
    Chain(
        chain_id=2340,
        name="Atleta Testnet",
        chain="atleta",
        network="testnet",
        currency=EVMCurrency(name="Atleta", symbol="ATLA", decimals=18),
    ),
    Chain(
        chain_id=842,
        name="Taraxa Testnet",
        chain="taraxa",
        network="testnet",
        currency=EVMCurrency(name="Taraxa", symbol="TARA", decimals=18),
    ),
    Chain(
        chain_id=808813,
        name="BOB Sepolia",
        chain="BOB",
        network="testnet",
        currency=EVMCurrency(name="Ether", symbol="ETH", decimals=18),
    ),
    Chain(
        chain_id=534352,
        name="Scroll",
        chain="scroll",
        network="mainnet",
        currency=EVMCurrency(name="Ether", symbol="ETH", decimals=18),
    ),
    Chain(
        chain_id=8453,
        name="Base",
        chain="base",
        network="mainnet",
        currency=EVMCurrency(name="Ether", symbol="ETH", decimals=18),
    ),
]


@dataclass
class ChainList(ConfigOptions):
    """returns chain information for the
    chain set in the telliot account

    Returns:
        currency, status...
    """

    chains: List[Chain] = field(default_factory=lambda: default_chain_list)

    def get_chain(self, chain: str = "ETH", network: str = "rinkeby") -> Optional[Chain]:
        """Get chain"""

        for ch in self.chains:
            if chain.lower() in ch.chain.lower():
                if network.lower() in ch.network.lower():
                    return ch

        return None



"""
Tests covering Pytelliot rpc connection  utils.
"""
import pytest
from brownie import chain
from requests.exceptions import ConnectionError
from requests.exceptions import HTTPError

from telliot_core.model.endpoints import EndpointList
from telliot_core.model.endpoints import RPCEndpoint

network = "mainnet"
provider = "pokt"


def test_rpc_endpoint():
    """RPCEndpoint connects to the blockchain"""
    url = "http://127.0.0.1:8545"  # local Ganache node
    endpt = RPCEndpoint(network=network, provider=provider, url=url)
    endpt.connect()
    chain.mine(10)
    assert endpt.web3.eth.block_number > 1

    print(endpt.get_state())


def test_very_bad_rpc_url():
    """an invalid url will raise an exception in RPCEndpoint"""
    url = "this is not a valid rpc url"
    endpt = RPCEndpoint(network=network, provider=provider, url=url)
    with pytest.raises(ValueError):
        _ = endpt.connect()


def test_incomplete_rpc_url():
    """an incomplete url will raise an exception in RPCEndpoint"""
    url = "https://eth-rpc.gateway.pokt.networ/"
    endpt = RPCEndpoint(network=network, provider=provider, url=url)
    try:
        _ = endpt.connect()
    except HTTPError:
        pass  # expected
    except ConnectionError:
        pass  # expected
    except ValueError as e:
        assert "Invalid request path" in str(e)
    else:
        pytest.fail("Expected ValueError or HTTPError.")


def test_endpoint_list():
    sl = EndpointList()
    # print(json.dumps(sl.get_state(), indent=2))
    ep11155111 = sl.find(chain_id=11155111)[0]
    assert ep11155111.network == "sepolia"
